﻿using System;
using System.Globalization;
using System.Runtime.InteropServices;

[Serializable, StructLayout(LayoutKind.Sequential), ComVisible(true)]
public struct Int24 : IComparable, IFormattable, IConvertible, IComparable<Int24>, IEquatable<Int24>
{
    private byte[] _value;

    public const int IntMaxValue = 8388607;
    public const int IntMinValue = -8388608;

    public static Int24 MaxValue { get { return new Int24(IntMaxValue); } }
    public static Int24 MinValue { get { return new Int24(IntMinValue); } }

    public static NumberFormatInfo CurrentInfo { get { throw new NotImplementedException(); } }

    public int Value
    {
        get
        {
            if (_value == null)
                _value = new byte[] { 0, 0, 0 };

            return BitConverter.ToInt32(new byte[] { _value[0], _value[1], _value[2], 0 }, 0);
        }
        set
        {
            if (value > IntMaxValue || value < IntMinValue)
                throw new OverflowException("Value was either too large or too small for an Int24.");

            byte[] bytes = BitConverter.GetBytes(value);
            _value = new[] { bytes[0], bytes[1], bytes[2] };
        }
    }

    public Int24(int value = 0) 
        : this()
    {
        if (value > IntMaxValue)
            value = value - 16777216;

        if (value > IntMaxValue || value < IntMinValue)
            throw new OverflowException("Value was either too large or too small for an Int24.");

        byte[] bytes = BitConverter.GetBytes(value);
        _value = new[] { bytes[0], bytes[1], bytes[2] };
    }

    public Int24(byte[] value)
        : this()
    {
        if (value == null || value.Length != 3)
            throw new OverflowException("An Int24 requires exactly 3 bytes.");

        _value = value;
    }

    public byte[] ToByteArray()
    {
        return _value;
    }

    public static Int24 FromByteArray(byte[] dataBytes)
    {
        return new Int24(dataBytes);
    }
    
    #region Operator Overloading

    public static implicit operator Int24(int value)
    {
        return new Int24(value);
    }

    public static implicit operator Int24(byte[] value)
    {
        return new Int24(value);
    }

    public static Int24 operator ~(Int24 value)
    {
        byte[] dataBytes = value.ToByteArray();
        return new Int24(new[]
                             {
                                 (byte) (~dataBytes[0]), (byte) (~dataBytes[1]), (byte) (~dataBytes[2])
                             });
    }

    public static Int24 operator +(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value + value2.Value);
    }

    public static Int24 operator -(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value - value2.Value);
    }

    public static Int24 operator *(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value*value2.Value);
    }

    public static Int24 operator /(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value/value2.Value);
    }

    public static Int24 operator %(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value%value2.Value);
    }

    public static Int24 operator ++(Int24 value)
    {
        value += 1;
        return value;
    }

    public static Int24 operator --(Int24 value)
    {
        value -= 1;
        return value;
    }

    public static bool operator ==(Int24 value1, Int24 value2)
    {
        return value1.Value == value2.Value;
    }

    public static bool operator !=(Int24 value1, Int24 value2)
    {
        return value1.Value != value2.Value;
    }

    public static bool operator <(Int24 value1, Int24 value2)
    {
        return value1.Value < value2.Value;
    }

    public static bool operator >(Int24 value1, Int24 value2)
    {
        return value1.Value > value2.Value;
    }

    public static bool operator <=(Int24 value1, Int24 value2)
    {
        return value1.Value <= value2.Value;
    }

    public static bool operator >=(Int24 value1, Int24 value2)
    {
        return value1.Value >= value2.Value;
    }

    public static Int24 operator &(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value & value2.Value);
    }

    public static Int24 operator |(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value | value2.Value);
    }

    public static Int24 operator ^(Int24 value1, Int24 value2)
    {
        return new Int24(value1.Value ^ value2.Value);
    }

    public static Int24 operator <<(Int24 value, int move)
    {
        int val = value.Value << move;

        if (val > IntMaxValue || val < IntMinValue)
        {
            val *= -1;
            if (val > IntMaxValue || val < IntMinValue)
                return new Int24();
        }

        return new Int24(val);
    }

    public static Int24 operator >>(Int24 value, int move)
    {
        int val = value.Value >> move;

        if (val > IntMaxValue || val < IntMinValue)
        {
            val *= -1;
            if (val > IntMaxValue || val < IntMinValue)
                return new Int24();
        }

        return new Int24(val);
    }

    #endregion

    //todo
    #region Parsing 

    public static bool TryParse(string s, out Int24 result)
    {
        throw new NotImplementedException();
    }

    public static bool TryParse(string s, NumberStyles style, IFormatProvider formatProvider, out Int24 result)
    {
        throw new NotImplementedException();
    }

    public static Int24 Parse(string s)
    {
        throw new NotImplementedException();
    }

    public static Int24 Parse(string s, IFormatProvider formatProvider)
    {
        throw new NotImplementedException();
    }

    public static Int24 Parse(string s, NumberStyles numberStyles)
    {
        throw new NotImplementedException();
    }

    public static Int24 Parse(string s, NumberStyles numberStyles, IFormatProvider formatProvider)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo 
    #region Convert from default types

    public static Int24 ConvertFromByte(byte value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromSByte(sbyte value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromShort(short value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromUShort(ushort value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromInt(int value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromUInt(uint value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromLong(long value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromULong(ulong value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromFloat(float value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromDouble(double value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromDecimal(decimal value)
    {
        throw new NotImplementedException();
    }

    public static Int24 ConvertFromString(string value)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo
    #region Implementation of IComparable

    public int CompareTo(object obj)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo
    #region Implementation of IFormattable

    public string ToString(string format, IFormatProvider formatProvider)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo
    #region Implementation of IConvertible

    public TypeCode GetTypeCode()
    {
        throw new NotImplementedException();
    }

    public bool ToBoolean(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public char ToChar(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public sbyte ToSByte(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public byte ToByte(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public short ToInt16(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public ushort ToUInt16(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public int ToInt32(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public uint ToUInt32(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public long ToInt64(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public ulong ToUInt64(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public float ToSingle(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public double ToDouble(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public decimal ToDecimal(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public DateTime ToDateTime(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public string ToString(IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    public object ToType(Type conversionType, IFormatProvider provider)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo
    #region Implementation of IComparable<Int24>

    public int CompareTo(Int24 other)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo
    #region Implementation of IEquatable<Int24>

    public bool Equals(Int24 other)
    {
        throw new NotImplementedException();
    }

    #endregion

    //todo
    #region Overrides from Base
    
    public override bool Equals(object obj)
    {
        throw new NotImplementedException();
    }

    public override int GetHashCode()
    {
        throw new NotImplementedException();
    }

    public override string ToString()
    {
        throw new NotImplementedException();
    }

    #endregion
}